@@ -18,6 +18,8 @@
 #include "parseconf.h"
 #include "oneprocess.h"
 #include "twoprocess.h"
+#include "standalone.h"
+#include "tcpwrap.h"
 
 /*
  * Forward decls of helper functions
@@ -39,19 +41,21 @@ main(int argc, const char* argv[])
     /* Login */
     1, INIT_MYSTR, INIT_MYSTR,
     /* Protocol state */
-    0, 1, INIT_MYSTR, 0,
+    0, 1, INIT_MYSTR, 0, 0,
     /* Session state */
     0,
     /* Userids */
-    -1, -1,
+    -1, -1, -1,
     /* Pre-chroot() cache */
-    INIT_MYSTR, INIT_MYSTR,
+    INIT_MYSTR, INIT_MYSTR, INIT_MYSTR, 1,
     /* Logging */
-    -1, INIT_MYSTR, 0, 0, 0, INIT_MYSTR, 0,
+    -1, -1, INIT_MYSTR, 0, 0, 0, INIT_MYSTR, 0,
     /* Buffers */
     INIT_MYSTR, INIT_MYSTR,
     /* Parent <-> child comms */
-    0, -1, -1
+    0, -1, -1,
+    /* Number of clients */
+    0, 0
   };
   int config_specified = 0;
   const char* p_config_name = VSFTP_DEFAULT_CONFIG;
@@ -71,6 +75,8 @@ main(int argc, const char* argv[])
     p_config_name = argv[1];
     config_specified = 1;
   }
+  /* Just get out unless we start with requisite privilege */
+  die_unless_privileged();
   /* This might need to open /dev/zero on systems lacking MAP_ANON. Needs
    * to be done early (i.e. before config file parse, which may use
    * anonymous pages
@@ -82,20 +88,30 @@ main(int argc, const char* argv[])
     int retval = vsf_sysutil_stat(p_config_name, &p_statbuf);
     if (!vsf_sysutil_retval_is_error(retval))
     {
-      vsf_parseconf_load_file(p_config_name);
+      vsf_parseconf_load_file(p_config_name, 1);
     }
     else if (config_specified)
     {
-      die("vsftpd: cannot open specified config file");
+      die2("vsftpd: cannot open config file:", p_config_name);
     }
     vsf_sysutil_free(p_statbuf);
   }
+  if (tunable_setproctitle_enable)
+  {
+    /* Warning -- warning -- may nuke argv, environ */
+    vsf_sysutil_setproctitle_init(argc, argv);
+  }
+  if (tunable_listen || tunable_listen_ipv6)
+  {
+    /* Standalone mode */
+    struct vsf_client_launch ret = vsf_standalone_main();
+    the_session.num_clients = ret.num_children;
+    the_session.num_this_ip = ret.num_this_ip;
+  }
   /* Sanity checks - exit with a graceful error message if our STDIN is not
    * a socket. Also check various config options don't collide.
    */
   do_sanity_checks();
-  /* Just get out unless we start with requisite privilege */
-  die_unless_privileged();
   /* Initializes session globals - e.g. IP addr's etc. */
   session_init(&the_session);
   /* Set up "environment", e.g. process group etc. */
@@ -105,26 +121,44 @@ main(int argc, const char* argv[])
    */
   vsf_log_init(&the_session);
   str_alloc_text(&the_session.remote_ip_str,
-                 vsf_sysutil_inet_ntoa(the_session.p_remote_addr));
+                 vsf_sysutil_inet_ntop(the_session.p_remote_addr));
   /* Set up options on the command socket */
   vsf_cmdio_sock_setup();
   if (tunable_setproctitle_enable)
   {
-    /* Warning -- warning -- may nuke argv, environ */
-    vsf_sysutil_setproctitle_init(argc, argv);
     vsf_sysutil_set_proctitle_prefix(&the_session.remote_ip_str);
     vsf_sysutil_setproctitle("connected");
   }
   /* We might chroot() very soon (one process model), so we need to open
    * any required config files here.
    */
+  if (tunable_tcp_wrappers)
+  {
+    the_session.tcp_wrapper_ok = vsf_tcp_wrapper_ok(VSFTP_COMMAND_FD);
+  }
+  {
+    const char* p_load_conf = vsf_sysutil_getenv("VSFTPD_LOAD_CONF");
+    if (p_load_conf)
+    {
+      vsf_parseconf_load_file(p_load_conf, 1);
+    }
+  }
   if (tunable_deny_email_enable)
   {
     int retval = str_fileread(&the_session.banned_email_str,
                               tunable_banned_email_file, VSFTP_CONF_FILE_MAX);
     if (vsf_sysutil_retval_is_error(retval))
     {
-      die("cannot open banned e-mail list file");
+      die2("cannot open banned e-mail list file:", tunable_banned_email_file);
+    }
+  }
+  if (tunable_banner_file)
+  {
+    int retval = str_fileread(&the_session.banner_str, tunable_banner_file,
+                              VSFTP_CONF_FILE_MAX);
+    if (vsf_sysutil_retval_is_error(retval))
+    {
+      die2("cannot open banner file:", tunable_banner_file);
     }
   }
   /* Special case - can force one process model if we've got a setup
@@ -173,11 +207,11 @@ do_sanity_checks(void)
   {
     if (tunable_local_enable)
     {
-      die("vsftpd: security: 'tunable_one_process_model' is anonymous only");
+      die("vsftpd: security: 'one_process_model' is anonymous only");
     }
     if (!vsf_sysdep_has_capabilities_as_non_root())
     {
-      die("vsftpd: security: 'tunable_one_process_model' needs a better OS");
+      die("vsftpd: security: 'one_process_model' needs a better OS");
     }
   }
   if (!tunable_local_enable && !tunable_anonymous_enable)
@@ -211,19 +245,32 @@ session_init(struct vsf_session* p_sess)
       vsf_sysutil_getpwnam(tunable_ftp_username);
     if (p_user == 0)
     {
-      die("vsftpd: cannot locate user specified in 'tunable_ftp_username'");
+      die2("vsftpd: cannot locate user specified in 'ftp_username':",
+           tunable_ftp_username);
     }
     p_sess->anon_ftp_uid = vsf_sysutil_user_getuid(p_user);
-
-    if (tunable_chown_uploads)
+  }
+  if (tunable_guest_enable)
+  {
+    const struct vsf_sysutil_user* p_user =
+      vsf_sysutil_getpwnam(tunable_guest_username);
+    if (p_user == 0)
     {
-      p_user = vsf_sysutil_getpwnam(tunable_chown_username);
-      if (p_user == 0)
-      {
-        die("vsf_sysutil_getpwnam");
-      }
-      p_sess->anon_upload_chown_uid = vsf_sysutil_user_getuid(p_user);
+      die2("vsftpd: cannot locate user specified in 'guest_username':",
+           tunable_guest_username);
     }
+    p_sess->guest_user_uid = vsf_sysutil_user_getuid(p_user);
+  }
+  if (tunable_chown_uploads)
+  {
+    const struct vsf_sysutil_user* p_user =
+      vsf_sysutil_getpwnam(tunable_chown_username);
+    if (p_user == 0)
+    {
+      die2("vsftpd: cannot locate user specified in 'chown_username':",
+           tunable_chown_username);
+    }
+    p_sess->anon_upload_chown_uid = vsf_sysutil_user_getuid(p_user);
   }
 }
 
