@@ -16,9 +16,15 @@
 #include "sysutil.h"
 #include "utility.h"
 
+static const char* s_p_saved_filename;
+static int s_strings_copied;
+
 /* File local functions */
 static void handle_config_setting(struct mystr* p_setting_str,
-                                  struct mystr* p_value_str);
+                                  struct mystr* p_value_str,
+                                  int errs_fatal);
+
+static void copy_string_settings(void);
 
 /* Tables mapping setting names to runtime variables */
 /* Boolean settings */
@@ -58,6 +64,25 @@ parseconf_bool_array[] =
   { "guest_enable", &tunable_guest_enable },
   { "userlist_enable", &tunable_userlist_enable },
   { "userlist_deny", &tunable_userlist_deny },
+  { "use_localtime", &tunable_use_localtime },
+  { "check_shell", &tunable_check_shell },
+  { "hide_ids", &tunable_hide_ids },
+  { "listen", &tunable_listen },
+  { "port_promiscuous", &tunable_port_promiscuous },
+  { "passwd_chroot_enable", &tunable_passwd_chroot_enable },
+  { "no_anon_password", &tunable_no_anon_password },
+  { "tcp_wrappers", &tunable_tcp_wrappers },
+  { "use_sendfile", &tunable_use_sendfile },
+  { "force_dot_files", &tunable_force_dot_files },
+  { "listen_ipv6", &tunable_listen_ipv6 },
+  { "dual_log_enable", &tunable_dual_log_enable },
+  { "syslog_enable", &tunable_syslog_enable },
+  { "background", &tunable_background },
+  { "virtual_use_local_privs", &tunable_virtual_use_local_privs },
+  { "session_support", &tunable_session_support },
+  { "download_enable", &tunable_download_enable },
+  { "dirlist_enable", &tunable_dirlist_enable },
+  { "chmod_enable", &tunable_chmod_enable },
   { 0, 0 }
 };
 
@@ -79,6 +104,11 @@ parseconf_uint_array[] =
   { "pasv_max_port", &tunable_pasv_max_port },
   { "anon_max_rate", &tunable_anon_max_rate },
   { "local_max_rate", &tunable_local_max_rate },
+  { "listen_port", &tunable_listen_port },
+  { "max_clients", &tunable_max_clients },
+  { "file_open_mode", &tunable_file_open_mode },
+  { "max_per_ip", &tunable_max_per_ip },
+  { "trans_chunk_size", &tunable_trans_chunk_size },
   { 0, 0 }
 };
 
@@ -93,6 +123,7 @@ parseconf_str_array[] =
   { "ftp_username", &tunable_ftp_username },
   { "chown_username", &tunable_chown_username },
   { "xferlog_file", &tunable_xferlog_file },
+  { "vsftpd_log_file", &tunable_vsftpd_log_file },
   { "message_file", &tunable_message_file },
   { "nopriv_user", &tunable_nopriv_user },
   { "ftpd_banner", &tunable_ftpd_banner },
@@ -103,20 +134,59 @@ parseconf_str_array[] =
   { "userlist_file", &tunable_userlist_file },
   { "anon_root", &tunable_anon_root },
   { "local_root", &tunable_local_root },
+  { "banner_file", &tunable_banner_file },
+  { "pasv_address", &tunable_pasv_address },
+  { "listen_address", &tunable_listen_address },
+  { "user_config_dir", &tunable_user_config_dir },
+  { "listen_address6", &tunable_listen_address6 },
+  { "cmds_allowed", &tunable_cmds_allowed },
   { 0, 0 }
 };
 
 void
-vsf_parseconf_load_file(const char* p_filename)
+vsf_parseconf_load_file(const char* p_filename, int errs_fatal)
 {
   struct mystr config_file_str = INIT_MYSTR;
   struct mystr config_setting_str = INIT_MYSTR;
   struct mystr config_value_str = INIT_MYSTR;
   unsigned int str_pos = 0;
-  int retval = str_fileread(&config_file_str, p_filename, VSFTP_CONF_FILE_MAX);
+  int retval;
+  if (!p_filename)
+  {
+    p_filename = s_p_saved_filename;
+  }
+  else
+  {
+    if (s_p_saved_filename)
+    {
+      vsf_sysutil_free((char*)s_p_saved_filename);
+    }
+    s_p_saved_filename = vsf_sysutil_strdup(p_filename);
+  }
+  if (!p_filename)
+  {
+    bug("null filename in vsf_parseconf_load_file");
+  }
+  if (!s_strings_copied)
+  {
+    s_strings_copied = 1;
+    /* A minor hack to make sure all strings are malloc()'ed so we can free
+     * them at some later date. Specifically handles strings embedded in the
+     * binary.
+     */
+    copy_string_settings();
+  }
+  retval = str_fileread(&config_file_str, p_filename, VSFTP_CONF_FILE_MAX);
   if (vsf_sysutil_retval_is_error(retval))
   {
-    die("cannot open config file");
+    if (errs_fatal)
+    {
+      die2("cannot open config file:", p_filename);
+    }
+    else
+    {
+      return;
+    }
   }
   while (str_getline(&config_file_str, &config_setting_str, &str_pos))
   {
@@ -127,7 +197,7 @@ vsf_parseconf_load_file(const char* p_filename)
     }
     /* Split into name=value pair */
     str_split_char(&config_setting_str, &config_value_str, '=');
-    handle_config_setting(&config_setting_str, &config_value_str);
+    handle_config_setting(&config_setting_str, &config_value_str, errs_fatal);
   }
   str_free(&config_file_str);
   str_free(&config_setting_str);
@@ -135,11 +205,19 @@ vsf_parseconf_load_file(const char* p_filename)
 }
 
 static void
-handle_config_setting(struct mystr* p_setting_str, struct mystr* p_value_str)
+handle_config_setting(struct mystr* p_setting_str, struct mystr* p_value_str,
+                      int errs_fatal)
 {
   if (str_isempty(p_value_str))
   {
-    die("missing value in config file");
+    if (errs_fatal)
+    {
+      die2("missing value in config file for: ", str_getbuf(p_setting_str));
+    }
+    else
+    {
+      return;
+    }
   }
   /* Is it a boolean value? */
   {
@@ -162,9 +240,10 @@ handle_config_setting(struct mystr* p_setting_str, struct mystr* p_value_str)
         {
           *(p_bool_setting->p_variable) = 0;
         }
-        else
+        else if (errs_fatal)
         {
-          die("bad bool value in config file");
+          die2("bad bool value in config file for: ",
+               str_getbuf(p_setting_str));
         }
         return;
       }
@@ -202,12 +281,35 @@ handle_config_setting(struct mystr* p_setting_str, struct mystr* p_value_str)
       if (str_equal_text(p_setting_str, p_str_setting->p_setting_name))
       {
         /* Got it */
-        *(p_str_setting->p_variable) = str_strdup(p_value_str);
+        const char** p_curr_setting = p_str_setting->p_variable;
+        if (*p_curr_setting)
+        {
+          vsf_sysutil_free((char*)*p_curr_setting);
+        }
+        *p_curr_setting = str_strdup(p_value_str);
         return;
       }
       p_str_setting++;
     }
   }
-  die("unrecognised variable in config file");
+  if (errs_fatal)
+  {
+    die2("unrecognised variable in config file: ", str_getbuf(p_setting_str));
+  }
+}
+
+static void
+copy_string_settings(void)
+{
+  const struct parseconf_str_setting* p_str_setting = parseconf_str_array;
+  while (p_str_setting->p_setting_name != 0)
+  {
+    if (*p_str_setting->p_variable != 0)
+    {
+      *p_str_setting->p_variable =
+          vsf_sysutil_strdup(*p_str_setting->p_variable);
+    }
+    p_str_setting++;
+  }
 }
 
