@@ -18,6 +18,8 @@
 #include "sysdeputil.h"
 #include "sysutil.h"
 #include "session.h"
+#include "banner.h"
+#include "logging.h"
 
 /* Functions used */
 static void emit_greeting(struct vsf_session* p_sess);
@@ -43,10 +45,46 @@ init_connection(struct vsf_session* p_sess)
 static void
 emit_greeting(struct vsf_session* p_sess)
 {
-  if (tunable_ftpd_banner == 0)
+  struct mystr str_log_line = INIT_MYSTR;
+  /* Check for client limits (standalone mode only) */
+  if (tunable_max_clients > 0 &&
+      p_sess->num_clients > tunable_max_clients)
   {
-    vsf_cmdio_write(p_sess, FTP_GREET, "ready, dude (vsFTPd " VSF_VERSION 
-                    ": beat me, break me)");
+    str_alloc_text(&str_log_line, "Connection refused: too many sessions.");
+    vsf_log_line(p_sess, kVSFLogEntryConnection, &str_log_line);
+    vsf_cmdio_write_noblock(p_sess, FTP_TOO_MANY_USERS,
+                    "There are too many connected users, please try later.");
+    vsf_sysutil_exit(0);
+  }
+  if (tunable_max_per_ip > 0 &&
+      p_sess->num_this_ip > tunable_max_per_ip)
+  {
+    str_alloc_text(&str_log_line,
+                   "Connection refused: too many sessions for this address.");
+    vsf_log_line(p_sess, kVSFLogEntryConnection, &str_log_line);
+    vsf_cmdio_write_noblock(p_sess, FTP_IP_LIMIT,
+        "There are too many connections from your internet address.");
+    vsf_sysutil_exit(0);
+  }
+  if (!p_sess->tcp_wrapper_ok)
+  {
+    str_alloc_text(&str_log_line,
+                   "Connection refused: tcp_wrappers denial.");
+    vsf_log_line(p_sess, kVSFLogEntryConnection, &str_log_line);
+    vsf_cmdio_write_noblock(p_sess, FTP_IP_DENY, "Service not available.");
+    vsf_sysutil_exit(0);
+  }
+  vsf_log_line(p_sess, kVSFLogEntryConnection, &str_log_line);
+  if (!str_isempty(&p_sess->banner_str))
+  {
+    vsf_banner_write(p_sess, &p_sess->banner_str, FTP_GREET);
+    str_free(&p_sess->banner_str);
+    vsf_cmdio_write(p_sess, FTP_GREET, "");
+  }
+  else if (tunable_ftpd_banner == 0)
+  {
+    vsf_cmdio_write(p_sess, FTP_GREET, "(vsFTPd " VSF_VERSION 
+                    ")");
   }
   else
   {
@@ -89,11 +127,15 @@ handle_user_command(struct vsf_session* p_sess)
    * non-anonymous usernames in the hope we save passwords going plaintext
    * over the network
    */
+  int is_anon = 1;
   str_copy(&p_sess->user_str, &p_sess->ftp_arg_str);
   str_upper(&p_sess->ftp_arg_str);
-  if (!tunable_local_enable &&
-      !str_equal_text(&p_sess->ftp_arg_str, "FTP") &&
+  if (!str_equal_text(&p_sess->ftp_arg_str, "FTP") &&
       !str_equal_text(&p_sess->ftp_arg_str, "ANONYMOUS"))
+  {
+    is_anon = 0;
+  }
+  if (!tunable_local_enable && !is_anon)
   {
     vsf_cmdio_write(p_sess, FTP_LOGINERR,
                     "This FTP server is anonymous only.");
@@ -111,7 +153,16 @@ handle_user_command(struct vsf_session* p_sess)
       return;
     }
   }
-  vsf_cmdio_write(p_sess, FTP_GIVEPWORD, "Please specify the password.");
+  if (is_anon && tunable_no_anon_password)
+  {
+    /* Fake a password */
+    str_alloc_text(&p_sess->ftp_arg_str, "<no password>");
+    handle_pass_command(p_sess);
+  }
+  else
+  {
+    vsf_cmdio_write(p_sess, FTP_GIVEPWORD, "Please specify the password.");
+  }
 }
 
 static void
