@@ -28,6 +28,7 @@
 #include <sys/param.h>
 #include <sys/uio.h>
 
+
 /* Configuration.. here are the possibilities */
 #undef VSF_SYSDEP_HAVE_CAPABILITIES
 #undef VSF_SYSDEP_HAVE_SETKEEPCAPS
@@ -43,6 +44,10 @@
 #define VSF_SYSDEP_HAVE_SHADOW
 #define VSF_SYSDEP_HAVE_USERSHELL
 #define VSF_SYSDEP_HAVE_LIBCAP
+#define VSF_SYSDEP_HAVE_UTMPX
+
+#define __USE_GNU
+#include <utmpx.h>
 
 /* BEGIN config */
 #if defined(__linux__) && !defined(__ia64__) && !defined(__s390__)
@@ -85,11 +90,20 @@
 
 #ifdef __sgi
   #undef VSF_SYSDEP_HAVE_USERSHELL
+  #undef VSF_SYSDEP_HAVE_LIBCAP
 #endif
 
-#if (defined(__sgi) || defined(__hpux))
+#ifdef __osf__
+  #undef VSF_SYSDEP_HAVE_USERSHELL
+#endif
+
+#if (defined(__sgi) || defined(__hpux) || defined(__osf__))
   #define VSF_SYSDEP_NEED_OLD_FD_PASSING
 #endif
+
+#ifdef __sun
+  #define VSF_SYSDEP_HAVE_SOLARIS_SENDFILE
+#endif
 /* END config */
 
 /* PAM support - we include our own dummy version if the system lacks this */
@@ -101,6 +115,7 @@
 #include <shadow.h>
 #include <pwd.h>
 #include <unistd.h>
+#include <crypt.h>
 #endif
 
 /* Prefer libcap based capabilities over raw syscall capabilities */
@@ -116,7 +131,8 @@ _syscall2(int, capset, cap_user_header_t, header, const cap_user_data_t, data)
 #undef __FDMASK
 #endif /* VSF_SYSDEP_HAVE_CAPABILITIES */
 
-#ifdef VSF_SYSDEP_HAVE_LINUX_SENDFILE
+#if defined(VSF_SYSDEP_HAVE_LINUX_SENDFILE) || \
+    defined(VSF_SYSDEP_HAVE_SOLARIS_SENDFILE)
 #include <sys/sendfile.h>
 #elif defined(VSF_SYSDEP_HAVE_FREEBSD_SENDFILE)
 #include <sys/types.h>
@@ -148,9 +164,12 @@ static int s_zero_fd = -1;
 
 /* File private functions/variables */
 static int do_sendfile(const int out_fd, const int in_fd,
-                       long* p_offset, unsigned int num_send);
+                       unsigned int num_send, filesize_t start_pos);
 static void vsf_sysutil_setproctitle_internal(const char* p_text);
 static struct mystr s_proctitle_prefix_str;
+static void vsf_insert_uwtmp(const struct mystr* p_user_str,
+                             const struct mystr* p_host_str);
+static void vsf_remove_uwtmp(void);
 
 #ifndef VSF_SYSDEP_HAVE_PAM
 int
@@ -158,7 +177,6 @@ vsf_sysdep_check_auth(const struct mystr* p_user_str,
                       const struct mystr* p_pass_str,
                       const struct mystr* p_remote_host)
 {
-  const char* p_shell;
   const char* p_crypted;
   const struct passwd* p_pwd = getpwnam(str_getbuf(p_user_str));
   (void) p_remote_host;
@@ -167,17 +185,21 @@ vsf_sysdep_check_auth(const struct mystr* p_user_str,
     return 0;
   }
   #ifdef VSF_SYSDEP_HAVE_USERSHELL
-  while ((p_shell = getusershell()) != NULL)
+  if (tunable_check_shell)
   {
-    if (!vsf_sysutil_strcmp(p_shell, p_pwd->pw_shell))
+    const char* p_shell;
+    while ((p_shell = getusershell()) != NULL)
     {
-      break;
+      if (!vsf_sysutil_strcmp(p_shell, p_pwd->pw_shell))
+      {
+        break;
+      }
+    }
+    endusershell();
+    if (p_shell == NULL)
+    {
+      return 0;
     }
-  }
-  endusershell();
-  if (p_shell == NULL)
-  {
-    return 0;
   }
   #endif
   #ifdef VSF_SYSDEP_HAVE_SHADOW
@@ -217,9 +239,11 @@ vsf_sysdep_check_auth(const struct mystr* p_user_str,
 
 #else /* VSF_SYSDEP_HAVE_PAM */
 
+static pam_handle_t* s_pamh;
 static struct mystr s_pword_str;
 static int pam_conv_func(int nmsg, const struct pam_message** p_msg,
                          struct pam_response** p_reply, void* p_addata);
+static void vsf_auth_shutdown(void);
 
 int
 vsf_sysdep_check_auth(const struct mystr* p_user_str,
@@ -227,55 +251,93 @@ vsf_sysdep_check_auth(const struct mystr* p_user_str,
                       const struct mystr* p_remote_host)
 {
   int retval;
-  pam_handle_t* pamh = 0;
   struct pam_conv the_conv =
   {
     &pam_conv_func,
     0
   };
+  if (s_pamh != 0)
+  {
+    bug("vsf_sysdep_check_auth");
+  }
   str_copy(&s_pword_str, p_pass_str);
   retval = pam_start(tunable_pam_service_name,
-                     str_getbuf(p_user_str), &the_conv, &pamh);
+                     str_getbuf(p_user_str), &the_conv, &s_pamh);
   if (retval != PAM_SUCCESS)
   {
-    pam_end(pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  #ifdef PAM_RHOST
-  retval = pam_set_item(pamh, PAM_RHOST, str_getbuf(p_remote_host));
+#ifdef PAM_RHOST
+  retval = pam_set_item(s_pamh, PAM_RHOST, str_getbuf(p_remote_host));
   if (retval != PAM_SUCCESS)
   {
-    pam_end(pamh, 0);
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  #endif
-  retval = pam_authenticate(pamh, 0);
+#endif
+  retval = pam_authenticate(s_pamh, 0);
   if (retval != PAM_SUCCESS)
   {
-    pam_end(pamh, 0);
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  retval = pam_acct_mgmt(pamh, 0);
+  retval = pam_acct_mgmt(s_pamh, 0);
   if (retval != PAM_SUCCESS)
   {
-    pam_end(pamh, 0);
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  retval = pam_setcred(pamh, PAM_ESTABLISH_CRED);
+  retval = pam_setcred(s_pamh, PAM_ESTABLISH_CRED);
   if (retval != PAM_SUCCESS)
   {
-    pam_end(pamh, 0);
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  retval = pam_end(pamh, PAM_SUCCESS);
+  if (!tunable_session_support)
+  {
+    /* You're in already! */
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
+    return 1;
+  }
+  /* Must do this BEFORE opening a session for pam_limits to count us */
+  vsf_insert_uwtmp(p_user_str, p_remote_host);
+  retval = pam_open_session(s_pamh, 0);
   if (retval != PAM_SUCCESS)
   {
+    vsf_remove_uwtmp();
+    (void) pam_setcred(s_pamh, PAM_DELETE_CRED);
+    (void) pam_end(s_pamh, 0);
+    s_pamh = 0;
     return 0;
   }
-  /* It worked, cool */
+  /* We MUST ensure the PAM session, utmp, wtmp etc. are cleaned up, however
+   * we exit.
+   */
+  vsf_sysutil_set_exit_func(vsf_auth_shutdown);
+  /* You're in dude */
   return 1;
 }
 
+static void
+vsf_auth_shutdown(void)
+{
+  if (s_pamh == 0)
+  {
+    bug("vsf_auth_shutdown");
+  }
+  (void) pam_close_session(s_pamh, 0);
+  (void) pam_setcred(s_pamh, PAM_DELETE_CRED);
+  (void) pam_end(s_pamh, 0);
+  s_pamh = 0;
+  vsf_remove_uwtmp();
+}
+
 static int
 pam_conv_func(int nmsg, const struct pam_message** p_msg,
               struct pam_response** p_reply, void* p_addata)
@@ -483,57 +545,56 @@ vsf_sysdep_adopt_capabilities(unsigned int caps)
 
 int
 vsf_sysutil_sendfile(const int out_fd, const int in_fd,
-                     unsigned long* p_offset, unsigned long num_send,
+                     filesize_t* p_offset, filesize_t num_send,
                      unsigned int max_chunk)
 {
   /* Grr - why is off_t signed? */
-  long real_offset = *p_offset;
-  if (real_offset < 0)
+  if (*p_offset < 0 || num_send < 0)
   {
-    die("invalid offset in vsf_sysutil_sendfile");
+    die("invalid offset or send count in vsf_sysutil_sendfile");
+  }
+  if (max_chunk == 0)
+  {
+    max_chunk = INT_MAX;
   }
   while (num_send > 0)
   {
     int retval;
     unsigned int send_this_time;
-    /* For 64-bit platforms */
-    if (num_send > INT_MAX)
+    if (num_send > max_chunk)
     {
-      send_this_time = INT_MAX;
+      send_this_time = max_chunk;
     }
     else
     {
       send_this_time = (unsigned int) num_send;
     }
-    if (max_chunk != 0 && send_this_time > max_chunk)
-    {
-      send_this_time = max_chunk;
-    }
-    retval = do_sendfile(out_fd, in_fd, &real_offset, send_this_time);
-    if (real_offset < 0)
-    {
-      die("invalid offset returned in vsf_sysutil_sendfile");
-    }
-    *p_offset = real_offset;
+    /* Keep input file position in line with sendfile() calls */
+    vsf_sysutil_lseek_to(in_fd, *p_offset);
+    retval = do_sendfile(out_fd, in_fd, send_this_time, *p_offset);
     if (vsf_sysutil_retval_is_error(retval) || retval == 0)
     {
       return retval;
     }
-    num_send -= (unsigned long) retval;
+    num_send -= retval;
+    *p_offset += retval;
   }
   return 0;
 }
 
 static int do_sendfile(const int out_fd, const int in_fd,
-                       long* p_offset, unsigned int num_send)
+                       unsigned int num_send, filesize_t start_pos)
 {
   /* Probably should one day be shared with instance in ftpdataio.c */
   static char* p_recvbuf;
   unsigned int total_written = 0;
   int retval;
+  (void) start_pos;
 #if defined(VSF_SYSDEP_HAVE_LINUX_SENDFILE) || \
     defined(VSF_SYSDEP_HAVE_FREEBSD_SENDFILE) || \
-    defined(VSF_SYSDEP_HAVE_HPUX_SENDFILE)
+    defined(VSF_SYSDEP_HAVE_HPUX_SENDFILE) || \
+    defined(VSF_SYSDEP_HAVE_SOLARIS_SENDFILE)
+  if (tunable_use_sendfile)
   {
     static int s_sendfile_checked;
     static int s_runtime_sendfile_works;
@@ -542,27 +603,39 @@ static int do_sendfile(const int out_fd, const int in_fd,
       do
       {
   #ifdef VSF_SYSDEP_HAVE_LINUX_SENDFILE
-        retval = sendfile(out_fd, in_fd, p_offset, num_send);
+        retval = sendfile(out_fd, in_fd, NULL, num_send);
   #elif defined(VSF_SYSDEP_HAVE_FREEBSD_SENDFILE)
         {
+          /* XXX - start_pos will truncate on 32-bit machines - can we
+           * say "start from current pos"?
+           */
           off_t written = 0;
-          retval = sendfile(in_fd, out_fd, *p_offset, num_send, NULL,
+          retval = sendfile(in_fd, out_fd, start_pos, num_send, NULL,
                             &written, 0);
           /* Translate to Linux-like retval */
           if (written > 0)
           {
             retval = (int) written;
-            *p_offset += retval;
+          }
+        }
+  #elif defined(VSF_SYSDEP_HAVE_SOLARIS_SENDFILE)
+        {
+          off_t written = 0;
+          struct sendfilevec the_vec;
+          vsf_sysutil_memclr(&the_vec, sizeof(the_vec));
+          the_vec.sfv_fd = in_fd;
+          the_vec.sfv_off = start_pos;
+          the_vec.sfv_len = num_send;
+          retval = sendfilev(out_fd, &the_vec, 1, &written);
+          /* Translate to Linux-like retval */
+          if (written > 0)
+          {
+            retval = (int) written;
           }
         }
   #else /* must be VSF_SYSDEP_HAVE_HPUX_SENDFILE */
         {
-          retval = sendfile(out_fd, in_fd, *p_offset, num_send, NULL, 0);
-          /* Translate to Linux-like retval */
-          if (retval > 0)
-          {
-            *p_offset += retval;
-          }
+          retval = sendfile(out_fd, in_fd, start_pos, num_send, NULL, 0);
         }
   #endif /* VSF_SYSDEP_HAVE_LINUX_SENDFILE */
         vsf_sysutil_check_pending_actions(kVSFSysUtilIO, retval, out_fd);
@@ -578,11 +651,21 @@ static int do_sendfile(const int out_fd, const int in_fd,
           s_runtime_sendfile_works = 1;
         }
       }
-      if (s_runtime_sendfile_works)
+      if (!vsf_sysutil_retval_is_error(retval))
       {
         return retval;
       }
-      /* Fall thru to normal implementation. We won't check again. */
+      if (s_runtime_sendfile_works &&
+          vsf_sysutil_get_error() != kVSFSysUtilErrINVAL &&
+          vsf_sysutil_get_error() != kVSFSysUtilErrOPNOTSUPP)
+      {
+        return retval;
+      }
+      /* Fall thru to normal implementation. We won't check again. NOTE -
+       * also falls through if sendfile() is OK but it returns EINVAL. For
+       * Linux this means the file was not page cache backed. Original
+       * complaint was trying to serve files from an NTFS filesystem!
+       */
     }
   }
 #endif /* VSF_SYSDEP_HAVE_LINUX_SENDFILE || VSF_SYSDEP_HAVE_FREEBSD_SENDFILE */
@@ -609,7 +692,6 @@ static int do_sendfile(const int out_fd, const int in_fd,
       return -1;
     }
     num_read = (unsigned int) retval;
-    *p_offset += num_read;
     retval = vsf_sysutil_write_loop(out_fd, p_recvbuf, num_read);
     if (retval < 0)
     {
@@ -949,3 +1031,87 @@ vsf_sysutil_recv_fd(int sock_fd)
 
 #endif /* !VSF_SYSDEP_NEED_OLD_FD_PASSING */
 
+#ifndef VSF_SYSDEP_HAVE_UTMPX
+
+static void
+vsf_insert_uwtmp(const struct mystr* p_user_str,
+                 const struct mystr* p_host_str)
+{
+  (void) p_user_str;
+  (void) p_host_str;
+}
+
+static void
+vsf_remove_uwtmp(void)
+{
+}
+
+#else /* !VSF_SYSDEP_HAVE_UTMPX */
+
+/* IMHO, the pam_unix module REALLY should be doing this in its SM component */
+/* Statics */
+static int s_uwtmp_inserted;
+static struct utmpx s_utent;
+
+static void
+vsf_insert_uwtmp(const struct mystr* p_user_str,
+                 const struct mystr* p_host_str)
+{
+  if (sizeof(s_utent.ut_line) < 16)
+  {
+    return;
+  }
+  if (s_uwtmp_inserted)
+  {
+    bug("vsf_insert_uwtmp");
+  }
+  {
+    struct mystr line_str = INIT_MYSTR;
+    str_alloc_text(&line_str, "vsftpd:");
+    str_append_ulong(&line_str, vsf_sysutil_getpid());
+    if (str_getlen(&line_str) >= sizeof(s_utent.ut_line))
+    {
+      str_free(&line_str);
+      return;
+    }
+    vsf_sysutil_strcpy(s_utent.ut_line, str_getbuf(&line_str),
+                       sizeof(s_utent.ut_line));
+    str_free(&line_str);
+  }
+  s_uwtmp_inserted = 1;
+  s_utent.ut_type = USER_PROCESS;
+  s_utent.ut_pid = vsf_sysutil_getpid();
+  vsf_sysutil_strcpy(s_utent.ut_user, str_getbuf(p_user_str),
+                     sizeof(s_utent.ut_user));
+  vsf_sysutil_strcpy(s_utent.ut_host, str_getbuf(p_host_str),
+                     sizeof(s_utent.ut_host));
+  vsf_sysutil_update_cached_time();
+  s_utent.ut_tv.tv_sec = vsf_sysutil_get_cached_time_sec();
+  setutxent();
+  (void) pututxline(&s_utent);
+  endutxent();
+  updwtmpx(WTMPX_FILE, &s_utent);
+}
+
+static void
+vsf_remove_uwtmp(void)
+{
+  if (!s_uwtmp_inserted)
+  {
+    return;
+  }
+  s_uwtmp_inserted = 0;
+  s_utent.ut_type = DEAD_PROCESS;
+  vsf_sysutil_memclr(s_utent.ut_user, sizeof(s_utent.ut_user));
+  vsf_sysutil_memclr(s_utent.ut_host, sizeof(s_utent.ut_host));
+  s_utent.ut_tv.tv_sec = 0;
+  setutxent();
+  (void) pututxline(&s_utent);
+  endutxent();
+  vsf_sysutil_update_cached_time();
+  s_utent.ut_tv.tv_sec = vsf_sysutil_get_cached_time_sec();
+  updwtmpx(WTMPX_FILE, &s_utent);
+}
+
+#endif /* !VSF_SYSDEP_HAVE_UTMPX */
+
