@@ -14,6 +14,10 @@
 #define PRIVATE_HANDS_OFF_exit_status exit_status
 #include "sysutil.h"
 #include "utility.h"
+#include "tunables.h"
+
+/* Activate 64-bit file support on Linux/32bit */
+#define _FILE_OFFSET_BITS 64
 
 /* For Linux, this adds nothing :-) */
 #include "port/porting_junk.h"
@@ -43,6 +47,7 @@
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
 #include <limits.h>
+#include <syslog.h>
 
 /* Private variables to this file */
 /* Current umask() */
@@ -51,6 +56,8 @@ static unsigned int s_current_umask;
 static struct timeval s_current_time;
 /* Current pid */
 static int s_current_pid = -1;
+/* Exit function */
+static exitfunc_t s_exit_func;
 
 /* Our internal signal handling implementation details */
 static struct vsf_sysutil_sig_details
@@ -65,6 +72,16 @@ static vsf_context_io_t s_io_handler;
 static void* s_p_io_handler_private;
 static int s_io_handler_running;
 
+struct vsf_sysutil_sockaddr
+{
+  union
+  {
+    struct sockaddr u_sockaddr;
+    struct sockaddr_in u_sockaddr_in;
+    struct sockaddr_in6 u_sockaddr_in6;
+  } u;
+};
+
 /* File locals */
 static void vsf_sysutil_common_sighandler(int signum);
 static int vsf_sysutil_translate_sig(const enum EVSFSysUtilSignal sig);
@@ -74,6 +91,8 @@ static int vsf_sysutil_translate_memprot(
 static int vsf_sysutil_translate_openmode(
   const enum EVSFSysUtilOpenMode mode);
 static void vsf_sysutil_alloc_statbuf(struct vsf_sysutil_statbuf** p_ptr);
+static void vsf_sysutil_sockaddr_alloc(
+  struct vsf_sysutil_sockaddr** p_sockptr);
 
 static void
 vsf_sysutil_common_sighandler(int signum)
@@ -148,6 +167,9 @@ vsf_sysutil_translate_sig(const enum EVSFSysUtilSignal sig)
     case kVSFSysUtilSigURG:
       realsig = SIGURG;
       break;
+    case kVSFSysUtilSigHUP:
+      realsig = SIGHUP;
+      break;
     default:
       bug("unknown signal in vsf_sysutil_translate_sig");
       break;
@@ -194,6 +216,7 @@ vsf_sysutil_install_async_sighandler(const enum EVSFSysUtilSignal sig,
   int realsig = vsf_sysutil_translate_sig(sig);
   s_sig_details[realsig].p_private = NULL;
   s_sig_details[realsig].sync_sig_handler = NULL;
+  vsf_sysutil_block_sig(sig);
   vsf_sysutil_set_sighandler(realsig, handler);
 }
 
@@ -204,10 +227,10 @@ vsf_sysutil_set_sighandler(int sig, void (*p_handlefunc)(int))
   struct sigaction sigact;
   vsf_sysutil_memclr(&sigact, sizeof(sigact));
   sigact.sa_handler = p_handlefunc;
-  retval = sigemptyset(&sigact.sa_mask);
+  retval = sigfillset(&sigact.sa_mask);
   if (retval != 0)
   {
-    die("sigemptyset");
+    die("sigfillset");
   }
   retval = sigaction(sig, &sigact, NULL);
   if (retval != 0)
@@ -216,6 +239,51 @@ vsf_sysutil_set_sighandler(int sig, void (*p_handlefunc)(int))
   }
 }
 
+void
+vsf_sysutil_block_sig(const enum EVSFSysUtilSignal sig)
+{
+  sigset_t sset;
+  int retval;
+  int realsig = vsf_sysutil_translate_sig(sig);
+  retval = sigemptyset(&sset);
+  if (retval != 0)
+  {
+    die("sigemptyset");
+  }
+  retval = sigaddset(&sset, realsig);
+  if (retval != 0)
+  {
+    die("sigaddset");
+  }
+  retval = sigprocmask(SIG_BLOCK, &sset, NULL);
+  if (retval != 0)
+  {
+    die("sigprocmask");
+  }
+}
+
+void
+vsf_sysutil_unblock_sig(const enum EVSFSysUtilSignal sig)
+{
+  sigset_t sset;
+  int retval;
+  int realsig = vsf_sysutil_translate_sig(sig);
+  retval = sigemptyset(&sset);
+  if (retval != 0)
+  {
+    die("sigemptyset");
+  }
+  retval = sigaddset(&sset, realsig);
+  if (retval != 0)
+  {
+    die("sigaddset");
+  }
+  retval = sigprocmask(SIG_UNBLOCK, &sset, NULL);
+  if (retval != 0)
+  {
+    die("sigprocmask");
+  }
+}
 void
 vsf_sysutil_install_io_handler(vsf_context_io_t handler, void* p_private)
 {
@@ -351,32 +419,30 @@ vsf_sysutil_write_loop(const int fd, const void* p_buf, unsigned int size)
   }
 }
 
-unsigned long
+filesize_t
 vsf_sysutil_get_file_offset(const int file_fd)
 {
-  off_t retval = lseek(file_fd, 0, SEEK_CUR);
+  filesize_t retval = lseek(file_fd, 0, SEEK_CUR);
   if (retval < 0)
   {
     die("lseek");
   }
-  return (unsigned long) retval;
+  return retval;
 }
 
-unsigned long
-vsf_sysutil_lseek_to(const int fd, unsigned long seek_pos)
+void
+vsf_sysutil_lseek_to(const int fd, filesize_t seek_pos)
 {
-  off_t retval;
-  off_t real_seek_pos = (off_t) seek_pos;
-  if (real_seek_pos < 0)
+  filesize_t retval;
+  if (seek_pos < 0)
   {
     die("negative seek_pos in vsf_sysutil_lseek_to");
   }
-  retval = lseek(fd, real_seek_pos, SEEK_SET);
+  retval = lseek(fd, seek_pos, SEEK_SET);
   if (retval < 0)
   {
     die("lseek");
   }
-  return (unsigned long) retval;
 }
 
 void*
@@ -435,21 +501,49 @@ vsf_sysutil_getpid(void)
 int
 vsf_sysutil_fork(void)
 {
-  int retval = fork();
+  /* Child does NOT inherit exit function */
+  exitfunc_t curr_func = s_exit_func;
+  int retval;
+  s_exit_func = 0;
+  retval = vsf_sysutil_fork_failok();
+  if (retval != 0)
+  {
+    s_exit_func = curr_func;
+  }
   if (retval < 0)
   {
     die("fork");
   }
-  else if (retval == 0)
+  return retval;
+}
+
+int
+vsf_sysutil_fork_failok(void)
+{
+  int retval = fork();
+  if (retval == 0)
   {
     s_current_pid = -1;
   }
   return retval;
 }
 
+void
+vsf_sysutil_set_exit_func(exitfunc_t exitfunc)
+{
+  s_exit_func = exitfunc;
+}
+
 void
 vsf_sysutil_exit(int exit_code)
 {
+  if (s_exit_func)
+  {
+    exitfunc_t curr_func = s_exit_func;
+    /* Prevent recursion */
+    s_exit_func = 0;
+    (*curr_func)();
+  }
   _exit(exit_code);
 }
 
@@ -471,6 +565,23 @@ vsf_sysutil_wait(void)
   }
 }
 
+int
+vsf_sysutil_wait_reap_one(void)
+{
+  int retval = waitpid(-1, NULL, WNOHANG);
+  if (retval == 0 || (retval < 0 && errno == ECHILD))
+  {
+    /* No more children */
+    return 0;
+  }
+  if (retval < 0)
+  {
+    die("waitpid");
+  }
+  /* Got one */
+  return retval;
+}
+
 int
 vsf_sysutil_wait_get_retval(const struct vsf_sysutil_wait_retval* p_waitret)
 {
@@ -576,18 +687,13 @@ vsf_sysutil_activate_linger(int fd)
 }
 
 void
-vsf_sysutil_deactivate_linger(int fd)
+vsf_sysutil_deactivate_linger_failok(int fd)
 {
-  int retval;
   struct linger the_linger;
   the_linger.l_onoff = 0;
   the_linger.l_linger = 0;
-  retval = setsockopt(fd, SOL_SOCKET, SO_LINGER, &the_linger,
-                      sizeof(the_linger));
-  if (retval != 0)
-  {
-    die("setsockopt");
-  }
+  (void) setsockopt(fd, SOL_SOCKET, SO_LINGER, &the_linger,
+                    sizeof(the_linger));
 }
 
 void
@@ -645,24 +751,66 @@ vsf_sysutil_atoi(const char* p_str)
   return atoi(p_str);
 }
 
-long
-vsf_sysutil_atol(const char* p_str)
+filesize_t
+vsf_sysutil_a_to_filesize_t(const char* p_str)
 {
-  return atol(p_str);
+  /* atoll() is C99 standard - but even modern FreeBSD, OpenBSD don't have
+   * it, so we'll supply our own
+   */
+  filesize_t result = 0;
+  filesize_t mult = 1;
+  unsigned int len = vsf_sysutil_strlen(p_str);
+  unsigned int i;
+  /* Bail if the number is excessively big (petabytes!) */
+  if (len > 15)
+  {
+    return 0;
+  }
+  for (i=0; i<len; ++i)
+  {
+    char the_char = p_str[len-(i+1)];
+    filesize_t val;
+    if (the_char < '0' || the_char > '9')
+    {
+      return 0;
+    }
+    val = the_char - '0';
+    val *= mult;
+    result += val;
+    mult *= 10;
+  }
+  return result;
 }
 
 const char*
 vsf_sysutil_ulong_to_str(unsigned long the_ulong)
 {
-  static char int_buf[128];
-  (void) snprintf(int_buf, sizeof(int_buf), "%lu", the_ulong);
-  return int_buf;
+  static char ulong_buf[32];
+  (void) snprintf(ulong_buf, sizeof(ulong_buf), "%lu", the_ulong);
+  return ulong_buf;
+}
+
+const char*
+vsf_sysutil_filesize_t_to_str(filesize_t the_filesize)
+{
+  static char filesize_buf[32];
+  if (sizeof(long) == 8)
+  {
+    /* Avoid using non-standard %ll if we can */
+    (void) snprintf(filesize_buf, sizeof(filesize_buf), "%ld",
+                    (long) the_filesize);
+  }
+  else
+  {
+    (void) snprintf(filesize_buf, sizeof(filesize_buf), "%lld", the_filesize);
+  }
+  return filesize_buf;
 }
 
 const char*
 vsf_sysutil_double_to_str(double the_double)
 {
-  static char double_buf[128];
+  static char double_buf[32];
   (void) snprintf(double_buf, sizeof(double_buf), "%.2f", the_double);
   return double_buf;
 }
@@ -670,7 +818,7 @@ vsf_sysutil_double_to_str(double the_double)
 const char*
 vsf_sysutil_uint_to_octal(unsigned int the_uint)
 {
-  static char octal_buf[128];
+  static char octal_buf[32];
   if (the_uint == 0)
   {
     octal_buf[0] = '0';
@@ -691,18 +839,19 @@ vsf_sysutil_octal_to_uint(const char* p_str)
   int seen_non_zero_digit = 0;
   while (*p_str != '\0')
   {
-    if (!isdigit(*p_str) || *p_str > '7')
+    int digit = *p_str;
+    if (!isdigit(digit) || digit > '7')
     {
       break;
     }
-    if (*p_str != '0')
+    if (digit != '0')
     {
       seen_non_zero_digit = 1;
     }
     if (seen_non_zero_digit)
     {
       result <<= 3;
-      result += ( *p_str - '0' );
+      result += (digit - '0');
     }
     p_str++;
   }
@@ -727,16 +876,16 @@ vsf_sysutil_isprint(int the_char)
   /* From Solar - we know better than some libc's! Don't let any potential
    * control chars through
    */
-  unsigned char uchar = (unsigned char) the_char;
-  if (uchar <= 31)
+  unsigned char uc = (unsigned char) the_char;
+  if (uc <= 31)
   {
     return 0;
   }
-  if (uchar == 177)
+  if (uc == 177)
   {
     return 0;
   }
-  if (uchar >= 128 && uchar <= 159)
+  if (uc >= 128 && uc <= 159)
   {
     return 0;
   }
@@ -844,6 +993,17 @@ vsf_sysutil_memcpy(void* p_dest, const void* p_src, const unsigned int size)
   memcpy(p_dest, p_src, size);
 }
 
+void
+vsf_sysutil_strcpy(char* p_dest, const char* p_src, unsigned int maxsize)
+{
+  if (maxsize == 0)
+  {
+    return;
+  }
+  strncpy(p_dest, p_src, maxsize);
+  p_dest[maxsize - 1] = '\0';
+}
+
 int
 vsf_sysutil_memcmp(const void* p_src1, const void* p_src2, unsigned int size)
 {
@@ -943,26 +1103,44 @@ int
 vsf_sysutil_open_file(const char* p_filename,
                       const enum EVSFSysUtilOpenMode mode)
 {
-  return open(p_filename, vsf_sysutil_translate_openmode(mode));
+  return open(p_filename, vsf_sysutil_translate_openmode(mode) | O_NONBLOCK);
 }
 
 int
 vsf_sysutil_create_file(const char* p_filename)
 {
-  /* NOTE! 0666 mode used so umask() decides end mode */
-  return open(p_filename, O_CREAT | O_EXCL | O_WRONLY | O_APPEND, 0666);
+  /* umask() also contributes to end mode */
+  return open(p_filename, O_CREAT | O_EXCL | O_WRONLY | O_APPEND,
+              tunable_file_open_mode);
 }
 
 int
 vsf_sysutil_create_overwrite_file(const char* p_filename)
 {
-  return open(p_filename, O_CREAT | O_TRUNC | O_WRONLY | O_APPEND, 0666);
+  return open(p_filename, O_CREAT | O_TRUNC | O_WRONLY |
+                          O_APPEND | O_NONBLOCK,
+              tunable_file_open_mode);
 }
 
 int
 vsf_sysutil_create_or_open_file(const char* p_filename, unsigned int mode)
 {
-  return open(p_filename, O_CREAT | O_WRONLY | O_APPEND, mode);
+  return open(p_filename, O_CREAT | O_WRONLY | O_APPEND | O_NONBLOCK, mode);
+}
+
+void
+vsf_sysutil_dupfd2(int old_fd, int new_fd)
+{
+  int retval;
+  if (old_fd == new_fd)
+  {
+    return;
+  }
+  retval = dup2(old_fd, new_fd);
+  if (retval != new_fd)
+  {
+    die("dup2");
+  }
 }
 
 void
@@ -1112,15 +1290,23 @@ vsf_sysutil_statbuf_get_perms(const struct vsf_sysutil_statbuf* p_statbuf)
 }
 
 const char*
-vsf_sysutil_statbuf_get_date(const struct vsf_sysutil_statbuf* p_statbuf)
+vsf_sysutil_statbuf_get_date(const struct vsf_sysutil_statbuf* p_statbuf,
+                             int use_localtime)
 {
-  static char datebuf[128];
+  static char datebuf[64];
   int retval;
   struct tm* p_tm;
   const struct stat* p_stat = (const struct stat*) p_statbuf;
   long local_time = vsf_sysutil_get_cached_time_sec();
   const char* p_date_format = "%b %d %H:%M";
-  p_tm = gmtime(&p_stat->st_mtime);
+  if (!use_localtime)
+  {
+    p_tm = gmtime(&p_stat->st_mtime);
+  }
+  else
+  {
+    p_tm = localtime(&p_stat->st_mtime);
+  }
   /* Is this a future or 6 months old date? If so, we drop to year format */
   if (p_stat->st_mtime > local_time ||
       (local_time - p_stat->st_mtime) > 60*60*24*182)
@@ -1138,12 +1324,22 @@ vsf_sysutil_statbuf_get_date(const struct vsf_sysutil_statbuf* p_statbuf)
 
 const char*
 vsf_sysutil_statbuf_get_numeric_date(
-  const struct vsf_sysutil_statbuf* p_statbuf)
+  const struct vsf_sysutil_statbuf* p_statbuf,
+  int use_localtime)
 {
   static char datebuf[15];
   const struct stat* p_stat = (const struct stat*) p_statbuf;
-  struct tm* p_tm = gmtime(&p_stat->st_mtime);
-  int retval = strftime(datebuf, sizeof(datebuf), "%Y%m%d%H%M%S", p_tm);
+  struct tm* p_tm;
+  int retval;
+  if (!use_localtime)
+  {
+    p_tm = gmtime(&p_stat->st_mtime);
+  }
+  else
+  {
+    p_tm = localtime(&p_stat->st_mtime);
+  }
+  retval = strftime(datebuf, sizeof(datebuf), "%Y%m%d%H%M%S", p_tm);
   if (retval == 0)
   {
     die("strftime");
@@ -1151,7 +1347,7 @@ vsf_sysutil_statbuf_get_numeric_date(
   return datebuf;
 }
 
-unsigned long
+filesize_t
 vsf_sysutil_statbuf_get_size(const struct vsf_sysutil_statbuf* p_statbuf)
 {
   const struct stat* p_stat = (const struct stat*) p_statbuf;
@@ -1159,7 +1355,7 @@ vsf_sysutil_statbuf_get_size(const struct vsf_sysutil_statbuf* p_statbuf)
   {
     die("invalid inode size in vsf_sysutil_statbuf_get_size");
   }
-  return (unsigned long) p_stat->st_size;
+  return p_stat->st_size;
 }
 
 int
@@ -1218,6 +1414,16 @@ vsf_sysutil_fchown(const int fd, const int uid, const int gid)
   }
 }
 
+void
+vsf_sysutil_fchmod(const int fd, unsigned int mode)
+{
+  mode = mode & 0777;
+  if (fchmod(fd, mode))
+  {
+    die("fchmod");
+  }
+}
+
 int
 vsf_sysutil_chmod(const char* p_filename, unsigned int mode)
 {
@@ -1300,6 +1506,12 @@ vsf_sysutil_get_error(void)
     case EINTR:
       retval = kVSFSysUtilErrINTR;
       break;
+    case EINVAL:
+      retval = kVSFSysUtilErrINVAL;
+      break;
+    case EOPNOTSUPP:
+      retval = kVSFSysUtilErrOPNOTSUPP;
+      break;
   }
   return retval;
 }
@@ -1315,7 +1527,18 @@ vsf_sysutil_get_ipv4_sock(void)
   return retval;
 }
 
-const struct vsf_sysutil_socketpair_retval
+int
+vsf_sysutil_get_ipv6_sock(void)
+{
+  int retval = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
+  if (retval < 0)
+  {
+    die("socket");
+  }
+  return retval;
+}
+
+struct vsf_sysutil_socketpair_retval
 vsf_sysutil_unix_dgram_socketpair(void)
 {
   struct vsf_sysutil_socketpair_retval retval;
@@ -1333,8 +1556,21 @@ vsf_sysutil_unix_dgram_socketpair(void)
 int
 vsf_sysutil_bind(int fd, const struct vsf_sysutil_sockaddr* p_sockptr)
 {
-  struct sockaddr* p_sockaddr = (struct sockaddr*) p_sockptr;
-  return bind(fd, p_sockaddr, sizeof(struct sockaddr_in));
+  const struct sockaddr* p_sockaddr = &p_sockptr->u.u_sockaddr;
+  int len = 0;
+  if (p_sockaddr->sa_family == AF_INET)
+  {
+    len = sizeof(struct sockaddr_in);
+  }
+  else if (p_sockaddr->sa_family == AF_INET6)
+  {
+    len = sizeof(struct sockaddr_in6);
+  }
+  else
+  {
+    die("can only support ipv4 and ipv6 currently");
+  }
+  return bind(fd, p_sockaddr, len);
 }
 
 void
@@ -1351,12 +1587,15 @@ int
 vsf_sysutil_accept_timeout(int fd, struct vsf_sysutil_sockaddr** p_sockptr,
                            unsigned int wait_seconds)
 {
-  struct sockaddr_in remote_addr;
+  struct vsf_sysutil_sockaddr remote_addr;
   int retval;
   fd_set accept_fdset;
   struct timeval timeout;
   unsigned int socklen = sizeof(remote_addr);
-  vsf_sysutil_sockaddr_clear(p_sockptr);
+  if (p_sockptr)
+  {
+    vsf_sysutil_sockaddr_clear(p_sockptr);
+  }
   if (wait_seconds > 0)
   {
     FD_ZERO(&accept_fdset);
@@ -1374,7 +1613,8 @@ vsf_sysutil_accept_timeout(int fd, struct vsf_sysutil_sockaddr** p_sockptr,
       return -1;
     }
   }
-  retval = accept(fd, (struct sockaddr*) &remote_addr, &socklen);
+  retval = accept(fd, &remote_addr.u.u_sockaddr, &socklen);
+  vsf_sysutil_check_pending_actions(kVSFSysUtilUnknown, 0, 0);
   if (retval < 0)
   {
     return retval;
@@ -1384,12 +1624,27 @@ vsf_sysutil_accept_timeout(int fd, struct vsf_sysutil_sockaddr** p_sockptr,
   {
     return -1;
   }
-  if (remote_addr.sin_family != AF_INET)
+  if (remote_addr.u.u_sockaddr.sa_family != AF_INET &&
+      remote_addr.u.u_sockaddr.sa_family != AF_INET6)
   {
-    die("can only support ipv4 currently");
+    die("can only support ipv4 and ipv6 currently");
+  }
+  if (p_sockptr)
+  {
+    vsf_sysutil_sockaddr_alloc(p_sockptr);
+    if (remote_addr.u.u_sockaddr.sa_family == AF_INET)
+    {
+      vsf_sysutil_memclr(&remote_addr.u.u_sockaddr_in.sin_zero,
+                         sizeof(remote_addr.u.u_sockaddr_in.sin_zero));
+      vsf_sysutil_memcpy(*p_sockptr, &remote_addr.u.u_sockaddr_in,
+                         sizeof(remote_addr.u.u_sockaddr_in));
+    }
+    else
+    {
+      vsf_sysutil_memcpy(*p_sockptr, &remote_addr.u.u_sockaddr_in6,
+                         sizeof(remote_addr.u.u_sockaddr_in6));
+    }
   }
-  *p_sockptr = vsf_sysutil_malloc(sizeof(remote_addr));
-  vsf_sysutil_memcpy(*p_sockptr, &remote_addr, sizeof(remote_addr));
   return retval;
 }
 
@@ -1397,14 +1652,26 @@ int
 vsf_sysutil_connect_timeout(int fd, const struct vsf_sysutil_sockaddr* p_addr,
                             unsigned int wait_seconds)
 {
-  const struct sockaddr_in* p_sockaddr = (const struct sockaddr_in*) p_addr;
-  unsigned int addrlen = sizeof(*p_sockaddr);
+  const struct sockaddr* p_sockaddr = &p_addr->u.u_sockaddr;
+  unsigned int addrlen = 0;
   int retval;
+  if (p_sockaddr->sa_family == AF_INET)
+  {
+    addrlen = sizeof(p_addr->u.u_sockaddr_in);
+  }
+  else if (p_sockaddr->sa_family == AF_INET6)
+  {
+    addrlen = sizeof(p_addr->u.u_sockaddr_in6);
+  }
+  else
+  {
+    die("can only support ipv4 and ipv6 currently");
+  }
   if (wait_seconds > 0)
   {
     vsf_sysutil_activate_noblock(fd);
   }
-  retval = connect(fd, (const struct sockaddr*)p_sockaddr, addrlen);
+  retval = connect(fd, p_sockaddr, addrlen);
   if (retval < 0 && errno == EINPROGRESS)
   {
     fd_set connect_fdset;
@@ -1444,41 +1711,51 @@ vsf_sysutil_connect_timeout(int fd, const struct vsf_sysutil_sockaddr* p_addr,
 void
 vsf_sysutil_getsockname(int fd, struct vsf_sysutil_sockaddr** p_sockptr)
 {
-  struct sockaddr_in the_addr;
+  struct vsf_sysutil_sockaddr the_addr;
   int retval;
   unsigned int socklen = sizeof(the_addr);
   vsf_sysutil_sockaddr_clear(p_sockptr);
-  retval = getsockname(fd, (struct sockaddr*) &the_addr, &socklen);
+  retval = getsockname(fd, &the_addr.u.u_sockaddr, &socklen);
   if (retval != 0)
   {
     die("getsockname");
   }
-  if (the_addr.sin_family != AF_INET)
+  if (the_addr.u.u_sockaddr.sa_family != AF_INET &&
+      the_addr.u.u_sockaddr.sa_family != AF_INET6)
   {
-    die("can only support ipv4 currently");
+    die("can only support ipv4 and ipv6 currently");
   }
-  *p_sockptr = vsf_sysutil_malloc(sizeof(the_addr));
-  vsf_sysutil_memcpy(*p_sockptr, &the_addr, sizeof(the_addr));
+  vsf_sysutil_sockaddr_alloc(p_sockptr);
+  if (socklen > sizeof(the_addr))
+  {
+    socklen = sizeof(the_addr);
+  }
+  vsf_sysutil_memcpy(*p_sockptr, &the_addr, socklen);
 }
 
 void
 vsf_sysutil_getpeername(int fd, struct vsf_sysutil_sockaddr** p_sockptr)
 {
-  struct sockaddr_in the_addr;
+  struct vsf_sysutil_sockaddr the_addr;
   int retval;
   unsigned int socklen = sizeof(the_addr);
   vsf_sysutil_sockaddr_clear(p_sockptr);
-  retval = getpeername(fd, (struct sockaddr*) &the_addr, &socklen);
+  retval = getpeername(fd, &the_addr.u.u_sockaddr, &socklen);
   if (retval != 0)
   {
     die("getpeername");
   }
-  if (the_addr.sin_family != AF_INET)
+  if (the_addr.u.u_sockaddr.sa_family != AF_INET &&
+      the_addr.u.u_sockaddr.sa_family != AF_INET6)
   {
-    die("can only support ipv4 currently");
+    die("can only support ipv4 and ipv6 currently");
   }
-  *p_sockptr = vsf_sysutil_malloc(sizeof(the_addr));
-  vsf_sysutil_memcpy(*p_sockptr, &the_addr, sizeof(the_addr));
+  vsf_sysutil_sockaddr_alloc(p_sockptr);
+  if (socklen > sizeof(the_addr))
+  {
+    socklen = sizeof(the_addr);
+  }
+  vsf_sysutil_memcpy(*p_sockptr, &the_addr, socklen);
 }
 
 void
@@ -1501,69 +1778,264 @@ vsf_sysutil_sockaddr_clear(struct vsf_sysutil_sockaddr** p_sockptr)
   }
 }
 
+static void
+vsf_sysutil_sockaddr_alloc(struct vsf_sysutil_sockaddr** p_sockptr)
+{
+  vsf_sysutil_sockaddr_clear(p_sockptr);
+  *p_sockptr = vsf_sysutil_malloc(sizeof(**p_sockptr));
+  vsf_sysutil_memclr(*p_sockptr, sizeof(**p_sockptr));
+}
+
 void
 vsf_sysutil_sockaddr_alloc_ipv4(struct vsf_sysutil_sockaddr** p_sockptr)
 {
-  struct sockaddr_in new_addr;
-  vsf_sysutil_sockaddr_clear(p_sockptr);
-  *p_sockptr = vsf_sysutil_malloc(sizeof(new_addr));
-  vsf_sysutil_memclr(&new_addr, sizeof(new_addr));
-  new_addr.sin_family = AF_INET;
-  vsf_sysutil_memcpy(*p_sockptr, &new_addr, sizeof(new_addr));
+  vsf_sysutil_sockaddr_alloc(p_sockptr);
+  (*p_sockptr)->u.u_sockaddr.sa_family = AF_INET;
 }
 
 void
-vsf_sysutil_sockaddr_set_ipaddr(struct vsf_sysutil_sockaddr* p_sockptr,
-                                struct vsf_sysutil_ipv4addr the_addr)
+vsf_sysutil_sockaddr_alloc_ipv6(struct vsf_sysutil_sockaddr** p_sockptr)
 {
-  struct sockaddr_in* p_sockaddr = (struct sockaddr_in*) p_sockptr;
-  vsf_sysutil_memcpy(&p_sockaddr->sin_addr.s_addr, the_addr.data,
-                     sizeof(p_sockaddr->sin_addr.s_addr));
+  vsf_sysutil_sockaddr_alloc(p_sockptr);
+  (*p_sockptr)->u.u_sockaddr.sa_family = AF_INET6;
 }
 
-struct vsf_sysutil_ipv4addr
-vsf_sysutil_sockaddr_get_ipaddr(const struct vsf_sysutil_sockaddr* p_sockptr)
+void
+vsf_sysutil_sockaddr_clone(struct vsf_sysutil_sockaddr** p_sockptr,
+                           const struct vsf_sysutil_sockaddr* p_src)
 {
-  struct vsf_sysutil_ipv4addr retval;
-  const struct sockaddr_in* p_sockaddr = (const struct sockaddr_in*) p_sockptr;
-  vsf_sysutil_memcpy(retval.data, &p_sockaddr->sin_addr.s_addr,
-                     sizeof(retval.data));
-  return retval;
+  struct vsf_sysutil_sockaddr* p_sockaddr = 0;
+  vsf_sysutil_sockaddr_alloc(p_sockptr);
+  p_sockaddr = *p_sockptr;
+  if (p_src->u.u_sockaddr.sa_family == AF_INET)
+  {
+    p_sockaddr->u.u_sockaddr.sa_family = AF_INET;
+    vsf_sysutil_memcpy(&p_sockaddr->u.u_sockaddr_in.sin_addr,
+                       &p_src->u.u_sockaddr_in.sin_addr,
+                       sizeof(p_sockaddr->u.u_sockaddr_in.sin_addr));
+  }
+  else if (p_src->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    p_sockaddr->u.u_sockaddr.sa_family = AF_INET6;
+    vsf_sysutil_memcpy(&p_sockaddr->u.u_sockaddr_in6.sin6_addr,
+                       &p_src->u.u_sockaddr_in6.sin6_addr,
+                       sizeof(p_sockaddr->u.u_sockaddr_in6.sin6_addr));
+  }
+  else
+  {
+    die("can only support ipv4 and ipv6 currently");
+  }
 }
 
-struct vsf_sysutil_ipv4port
-vsf_sysutil_sockaddr_get_port(const struct vsf_sysutil_sockaddr* p_sockptr)
+int
+vsf_sysutil_sockaddr_addr_equal(const struct vsf_sysutil_sockaddr* p1,
+                                const struct vsf_sysutil_sockaddr* p2)
 {
-  struct vsf_sysutil_ipv4port retval;
-  const struct sockaddr_in* p_sockaddr = (const struct sockaddr_in*) p_sockptr;
-  vsf_sysutil_memcpy(retval.data, &p_sockaddr->sin_port, sizeof(retval.data));
-  return retval;
+  int family1 = p1->u.u_sockaddr.sa_family;
+  int family2 = p2->u.u_sockaddr.sa_family;
+  if (family1 != family2)
+  {
+    if (family1 == AF_INET && family2 == AF_INET6)
+    {
+      const void* p_ipv4_addr = vsf_sysutil_sockaddr_ipv6_v4(p2);
+      if (p_ipv4_addr &&
+          !vsf_sysutil_memcmp(p_ipv4_addr, &p1->u.u_sockaddr_in.sin_addr,
+                              sizeof(p1->u.u_sockaddr_in.sin_addr)))
+      {
+        return 1;
+      }
+    }
+    else if (family1 == AF_INET6 && family2 == AF_INET)
+    {
+      const void* p_ipv4_addr = vsf_sysutil_sockaddr_ipv6_v4(p1);
+      if (p_ipv4_addr &&
+          !vsf_sysutil_memcmp(p_ipv4_addr, &p2->u.u_sockaddr_in.sin_addr,
+                              sizeof(p2->u.u_sockaddr_in.sin_addr)))
+      {
+        return 1;
+      }
+    }
+    return 0;
+  }
+  if (family1 == AF_INET)
+  {
+    if (vsf_sysutil_memcmp(&p1->u.u_sockaddr_in.sin_addr,
+                           &p2->u.u_sockaddr_in.sin_addr,
+                           sizeof(p1->u.u_sockaddr_in.sin_addr)) == 0)
+    {
+      return 1;
+    }
+  }
+  else if (family1 == AF_INET6)
+  {
+    if (vsf_sysutil_memcmp(&p1->u.u_sockaddr_in6.sin6_addr,
+                           &p2->u.u_sockaddr_in6.sin6_addr,
+                           sizeof(p1->u.u_sockaddr_in6.sin6_addr)) == 0)
+    {
+      return 1;
+    }
+  }
+  return 0;
 }
 
-struct vsf_sysutil_ipv4port
-vsf_sysutil_ipv4port_from_int(unsigned int port)
+int
+vsf_sysutil_sockaddr_is_ipv6(const struct vsf_sysutil_sockaddr* p_sockaddr)
 {
-  struct vsf_sysutil_ipv4port retval;
-  unsigned short netorder_port = htons(port);
-  vsf_sysutil_memcpy(retval.data, &netorder_port, sizeof(retval.data));
-  return retval;
+  if (p_sockaddr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    return 1;
+  }
+  return 0;
+}
+
+int
+vsf_sysutil_sockaddr_same_family(const struct vsf_sysutil_sockaddr* p1,
+                                 const struct vsf_sysutil_sockaddr* p2)
+{
+  if (p1->u.u_sockaddr.sa_family == p2->u.u_sockaddr.sa_family)
+  {
+    return 1;
+  }
+  return 0;
+}
+
+void
+vsf_sysutil_sockaddr_set_ipv4addr(struct vsf_sysutil_sockaddr* p_sockptr,
+                                  const unsigned char* p_raw)
+{
+  if (p_sockptr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    vsf_sysutil_memcpy(&p_sockptr->u.u_sockaddr_in.sin_addr, p_raw,
+                       sizeof(p_sockptr->u.u_sockaddr_in.sin_addr));
+  }
+  else
+  {
+    bug("bad family");
+  }
+}
+
+void
+vsf_sysutil_sockaddr_set_ipv6addr(struct vsf_sysutil_sockaddr* p_sockptr,
+                                  const unsigned char* p_raw)
+{
+  if (p_sockptr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    vsf_sysutil_memcpy(&p_sockptr->u.u_sockaddr_in6.sin6_addr, p_raw,
+                       sizeof(p_sockptr->u.u_sockaddr_in6.sin6_addr));
+  }
+  else
+  {
+    bug("bad family");
+  }
+}
+
+const void*
+vsf_sysutil_sockaddr_ipv6_v4(const struct vsf_sysutil_sockaddr* p_addr)
+{
+  static char pattern[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
+  const unsigned char* p_addr_start;
+  if (p_addr->u.u_sockaddr.sa_family != AF_INET6)
+  {
+    return 0;
+  }
+  if (vsf_sysutil_memcmp(pattern, &p_addr->u.u_sockaddr_in6.sin6_addr, 12))
+  {
+    return 0;
+  }
+  p_addr_start = (const unsigned char*)&p_addr->u.u_sockaddr_in6.sin6_addr;
+  return &p_addr_start[12];
+}
+
+void*
+vsf_sysutil_sockaddr_get_raw_addr(struct vsf_sysutil_sockaddr* p_sockptr)
+{
+  if (p_sockptr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    return &p_sockptr->u.u_sockaddr_in.sin_addr;
+  }
+  else if (p_sockptr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    return &p_sockptr->u.u_sockaddr_in6.sin6_addr;
+  }
+  else
+  {
+    bug("bad family");
+  }
+  return 0;
+}
+
+unsigned int
+vsf_sysutil_get_ipaddr_size(void)
+{
+  struct vsf_sysutil_sockaddr addr;
+  unsigned int size = sizeof(addr.u.u_sockaddr_in.sin_addr);
+  unsigned int size2 = sizeof(addr.u.u_sockaddr_in6.sin6_addr);
+  if (size2 > size)
+  {
+    size = size2;
+  }
+  return size;
+}
+
+int
+vsf_sysutil_get_ipsock(const struct vsf_sysutil_sockaddr* p_addr)
+{
+  if (p_addr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    return vsf_sysutil_get_ipv4_sock();
+  }
+  else if (p_addr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    return vsf_sysutil_get_ipv6_sock();
+  }
+  else
+  {
+    bug("bad family");
+  }
+  return -1;
+}
+
+void
+vsf_sysutil_sockaddr_set_any(struct vsf_sysutil_sockaddr* p_sockaddr)
+{
+  if (p_sockaddr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    vsf_sysutil_memclr(&p_sockaddr->u.u_sockaddr_in.sin_addr,
+                       sizeof(p_sockaddr->u.u_sockaddr_in.sin_addr));
+  }
+  else if (p_sockaddr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    vsf_sysutil_memclr(&p_sockaddr->u.u_sockaddr_in6.sin6_addr,
+                       sizeof(p_sockaddr->u.u_sockaddr_in6.sin6_addr));
+  }
+  else
+  {
+    bug("bad family");
+  }
 }
 
 void
 vsf_sysutil_sockaddr_set_port(struct vsf_sysutil_sockaddr* p_sockptr,
-                              struct vsf_sysutil_ipv4port the_port)
+                              unsigned short the_port)
 {
-  struct sockaddr_in* p_sockaddrin = (struct sockaddr_in*) p_sockptr;
-  vsf_sysutil_memcpy(&p_sockaddrin->sin_port, the_port.data,
-                     sizeof(p_sockaddrin->sin_port));
+  if (p_sockptr->u.u_sockaddr.sa_family == AF_INET)
+  {
+    p_sockptr->u.u_sockaddr_in.sin_port = htons(the_port);
+  }
+  else if (p_sockptr->u.u_sockaddr.sa_family == AF_INET6)
+  {
+    p_sockptr->u.u_sockaddr_in6.sin6_port = htons(the_port);
+  }
+  else
+  {
+    bug("bad family");
+  }
 }
 
 int
-vsf_sysutil_is_port_reserved(const struct vsf_sysutil_ipv4port the_port)
+vsf_sysutil_is_port_reserved(unsigned short the_port)
 {
-  unsigned short netorder_port;
-  vsf_sysutil_memcpy(&netorder_port, the_port.data, sizeof(netorder_port));
-  if (ntohs(netorder_port) < IPPORT_RESERVED)
+  if (the_port < IPPORT_RESERVED)
   {
     return 1;
   }
@@ -1571,10 +2043,56 @@ vsf_sysutil_is_port_reserved(const struct vsf_sysutil_ipv4port the_port)
 }
 
 const char*
-vsf_sysutil_inet_ntoa(const struct vsf_sysutil_sockaddr* p_sockptr)
+vsf_sysutil_inet_ntop(const struct vsf_sysutil_sockaddr* p_sockptr)
 {
-  const struct sockaddr_in* p_sockaddr = (const struct sockaddr_in*) p_sockptr;
-  return inet_ntoa(p_sockaddr->sin_addr);
+  const struct sockaddr* p_sockaddr = &p_sockptr->u.u_sockaddr;
+  if (p_sockaddr->sa_family == AF_INET)
+  {
+    return inet_ntoa(p_sockptr->u.u_sockaddr_in.sin_addr);
+  }
+  else if (p_sockaddr->sa_family == AF_INET6)
+  {
+    static char inaddr_buf[64];
+    const char* p_ret = inet_ntop(AF_INET6,
+                                  &p_sockptr->u.u_sockaddr_in6.sin6_addr,
+                                  inaddr_buf, sizeof(inaddr_buf));
+    inaddr_buf[sizeof(inaddr_buf) - 1] = '\0';
+    if (p_ret == NULL)
+    {
+      inaddr_buf[0] = '\0';
+    }
+    return inaddr_buf;
+  }
+  else
+  {
+    die("can only support ipv4 and ipv6 currently");
+  }
+}
+
+const char*
+vsf_sysutil_inet_ntoa(const void* p_raw_addr)
+{
+  return inet_ntoa(*((struct in_addr*)p_raw_addr));
+}
+
+int
+vsf_sysutil_inet_aton(const char* p_text, struct vsf_sysutil_sockaddr* p_addr)
+{
+  struct in_addr sin_addr;
+  if (p_addr->u.u_sockaddr.sa_family != AF_INET)
+  {
+    bug("bad family");
+  }
+  if (inet_aton(p_text, &sin_addr))
+  {
+    vsf_sysutil_memcpy(&p_addr->u.u_sockaddr_in.sin_addr,
+                       &sin_addr, sizeof(p_addr->u.u_sockaddr_in.sin_addr));
+    return 1;
+  }
+  else
+  {
+    return 0;
+  }
 }
 
 struct vsf_sysutil_user*
@@ -1742,7 +2260,8 @@ vsf_sysutil_setegid(const struct vsf_sysutil_user* p_user)
 void
 vsf_sysutil_seteuid_numeric(int uid)
 {
-  int retval = seteuid(uid);
+  /* setreuid() would seem to be more portable than seteuid() */
+  int retval = setreuid(-1, uid);
   if (retval != 0)
   {
     die("seteuid");
@@ -1752,7 +2271,8 @@ vsf_sysutil_seteuid_numeric(int uid)
 void
 vsf_sysutil_setegid_numeric(int gid)
 {
-  int retval = setegid(gid);
+  /* setregid() would seem to be more portable than setegid() */
+  int retval = setregid(-1, gid);
   if (retval != 0)
   {
     die("setegid");
@@ -1824,17 +2344,31 @@ vsf_sysutil_tzset(void)
 const char*
 vsf_sysutil_get_current_date(void)
 {
-  static char datebuf[128];
+  static char datebuf[64];
   time_t curr_time;
   const struct tm* p_tm;
+  int i = 0;
   vsf_sysutil_update_cached_time();
   curr_time = vsf_sysutil_get_cached_time_sec();
   p_tm = localtime(&curr_time);
-  if (strftime(datebuf, sizeof(datebuf), "%a %b %d %H:%M:%S %Y", p_tm) == 0)
+  if (strftime(datebuf, sizeof(datebuf), "%a %b!%d %H:%M:%S %Y", p_tm) == 0)
   {
     die("strftime");
   }
   datebuf[sizeof(datebuf) - 1] = '\0';
+  /* This hack is because %e in strftime() isn't so portable */
+  while (datebuf[i] != '!' && datebuf[i] != '\0')
+  {
+    ++i;
+  }
+  if (datebuf[i] == '!')
+  {
+    datebuf[i] = ' ';
+    if (datebuf[i+1] == '0')
+    {
+      datebuf[i+1] = ' ';
+    }
+  }
   return datebuf;
 }
 
@@ -1884,3 +2418,30 @@ vsf_sysutil_sleep(double seconds)
   } while (retval == -1 && errno == EINTR);
 }
 
+char*
+vsf_sysutil_getenv(const char* p_var)
+{
+  return getenv(p_var);
+}
+
+void
+vsf_sysutil_openlog(void)
+{
+  int facility = LOG_DAEMON;
+#ifdef LOG_FTP
+  facility = LOG_FTP;
+#endif
+  openlog("vsftpd", LOG_NDELAY, facility);
+}
+
+void
+vsf_sysutil_syslog(const char* p_text, int severe)
+{
+  int prio = LOG_INFO;
+  if (severe)
+  {
+    prio = LOG_WARNING;
+  }
+  syslog(prio, "%s", p_text);
+}
+
