@@ -11,7 +11,7 @@ extern int tunable_chroot_local_user;         /* Restrict local to home dir */
 extern int tunable_write_enable;              /* Global enable writes */
 extern int tunable_anon_upload_enable;        /* Enable STOR for anon users */
 extern int tunable_anon_mkdir_write_enable;   /* MKD for anon */
-extern int tunable_anon_other_write_enable;   /* DELE RMD RNFR RNTO for anon */
+extern int tunable_anon_other_write_enable;   /* APPE DELE RMD RNFR for anon */
 extern int tunable_chown_uploads;             /* chown() anon uploaded files */
 extern int tunable_connect_from_port_20;      /* PORT connects from port 20 */
 extern int tunable_xferlog_enable;            /* Log transfers to a file */
@@ -32,6 +32,25 @@ extern int tunable_log_ftp_protocol;          /* Log FTP requests/responses */
 extern int tunable_guest_enable;              /* Remap guest users */
 extern int tunable_userlist_enable;           /* Explicit user allow or deny */
 extern int tunable_userlist_deny;             /* Is user list allow or deny? */
+extern int tunable_use_localtime;             /* Use local time or GMT? */
+extern int tunable_check_shell;               /* Use /etc/shells for non-PAM */
+extern int tunable_hide_ids;                  /* Show "ftp" in ls listings */
+extern int tunable_listen;                    /* Standalone (no inetd) mode? */
+extern int tunable_port_promiscuous;          /* Any any PORT connect IP */
+extern int tunable_passwd_chroot_enable;      /* chroot() based on passwd */
+extern int tunable_no_anon_password;          /* Do not ask for anon pword */
+extern int tunable_tcp_wrappers;              /* Standalone: do tcp wrappers */
+extern int tunable_use_sendfile;              /* Use sendfile() if we can */
+extern int tunable_force_dot_files;           /* Show dotfiles without -a */
+extern int tunable_listen_ipv6;               /* Standalone with IPv6 listen */
+extern int tunable_dual_log_enable;           /* Log vsftpd.log AND xferlog */
+extern int tunable_syslog_enable;             /* Use syslog not vsftpd.log */
+extern int tunable_background;                /* Background listener process */
+extern int tunable_virtual_use_local_privs;   /* Virtual user => local privs */
+extern int tunable_session_support;           /* utmp, wtmp, pam_session */
+extern int tunable_download_enable;           /* Can download anything? */
+extern int tunable_dirlist_enable;            /* Can see any dirs? */
+extern int tunable_chmod_enable;              /* Is CHMOD allowed? (local) */
 
 /* Integer/numeric defines */
 extern unsigned int tunable_accept_timeout;
@@ -45,12 +64,18 @@ extern unsigned int tunable_pasv_min_port;
 extern unsigned int tunable_pasv_max_port;
 extern unsigned int tunable_anon_max_rate;
 extern unsigned int tunable_local_max_rate;
+extern unsigned int tunable_listen_port;
+extern unsigned int tunable_max_clients;
+extern unsigned int tunable_file_open_mode;
+extern unsigned int tunable_max_per_ip;
+extern unsigned int tunable_trans_chunk_size;
 
 /* String defines */
 extern const char* tunable_secure_chroot_dir;
 extern const char* tunable_ftp_username;
 extern const char* tunable_chown_username;
 extern const char* tunable_xferlog_file;
+extern const char* tunable_vsftpd_log_file;
 extern const char* tunable_message_file;
 extern const char* tunable_nopriv_user;
 extern const char* tunable_ftpd_banner;
@@ -61,6 +86,12 @@ extern const char* tunable_guest_username;
 extern const char* tunable_userlist_file;
 extern const char* tunable_anon_root;
 extern const char* tunable_local_root;
+extern const char* tunable_banner_file;
+extern const char* tunable_pasv_address;
+extern const char* tunable_listen_address;
+extern const char* tunable_user_config_dir;
+extern const char* tunable_listen_address6;
+extern const char* tunable_cmds_allowed;
 
 #endif /* VSF_TUNABLES_H */
 
